"""
Utilities to automate RAW data downloading from
:ref:`http://www.cvlibs.net/datasets/kitti/raw_data.php`
"""

import logging
import os
import pickle
import re
import shutil
from multiprocessing.pool import ThreadPool
from typing import List
from zipfile import ZipFile

from pkg_resources import resource_filename
from tqdm import tqdm

from .._download import download_calib, download_file

__all__ = [
    "repos",
    "check_drives",
    "download",
]

logger = logging.getLogger(__name__)


def repos() -> List[str]:
    """
    Provides a plain list of urls pointing to all syncronized+rectified
    .zip files of KITTI dataset.
    These are also available at the address here_.

    .. _here: http://www.cvlibs.net/datasets/kitti/raw_data.php
    """

    drives_path = resource_filename(
        "torch_kitti.raw.synced_rectified.resources", "drives.pkl"
    )
    with open(drives_path, "rb") as f:
        drives = pickle.load(f)

    drive = re.compile(r"[0-9]+_[0-9]+_[0-9]+_drive_[0-9]+")

    repos = [drive.findall(d)[0] + "/" + d + ".zip" for d in drives]

    prefix = "https://s3.eu-central-1.amazonaws.com/avg-kitti/raw_data/"

    def concat_prefix(repo: str) -> str:
        return prefix + repo

    return list(map(concat_prefix, repos))


def check_drives(root: str) -> bool:
    """
    Performs a check of all drives contained in the kitti root folders
    and prints a report
    """

    date_match = re.compile("[0-9]+_[0-9]+_[0-9]+")
    drive_match = re.compile(r"[0-9]+_[0-9]+_[0-9]+_drive_[0-9]+_sync")

    missing = []
    for url in repos():
        drive = drive_match.findall(url)[0]
        date = date_match.findall(url)[0]
        if not os.path.exists(os.path.join(root, date, drive)):
            missing.append(drive)

    if len(missing) != 0:
        logging.error("missing drives: " + "\n".join("- " + drive for drive in missing))
        return False
    else:
        return True


def download(root_path: str, threads=6):
    """
    Provides a simple way to download and scaffold the whole kitti synced+rectified
    raw data in a folder from :ref:`http://www.cvlibs.net/datasets/kitti/raw_data.php`.
    The whole KITTI dataset contains about 170Gb.
    """

    date_match = re.compile("[0-9]+_[0-9]+_[0-9]+")
    drive_match = re.compile(r"[0-9]+_[0-9]+_[0-9]+_drive_[0-9]+_sync")

    # filter not existing repos
    all_repos = [
        url
        for url in repos()
        if not os.path.exists(
            os.path.join(
                root_path, date_match.findall(url)[0], drive_match.findall(url)[0]
            )
        )
    ]

    # chunk generator
    def chunks(lst, n):
        for i in range(0, len(lst), n):
            yield lst[i : i + n]

    bar = tqdm(total=len(all_repos), unit="zips", desc="download repos")

    def download_unzip(url):
        # params
        date = date_match.findall(url)[0]
        drive = drive_match.findall(url)[0]

        # create date folder if not exist
        os.makedirs(os.path.join(root_path, date), exist_ok=True)

        # download
        download_file(url, os.path.join(root_path, date, drive + ".zip"))

        # unzip
        with ZipFile(os.path.join(root_path, date, drive + ".zip"), "r") as zip_ref:
            zip_ref.extractall(os.path.join(root_path, date, drive + "_tmp"))

        # move folders
        os.rename(
            os.path.join(root_path, date, drive + "_tmp", date, drive),
            os.path.join(root_path, date, drive),
        )

        # remove old folder and zip file
        shutil.rmtree(os.path.join(root_path, date, drive + "_tmp"))
        os.remove(os.path.join(root_path, date, drive + ".zip"))

        # update bar
        bar.update(1)

    try:
        # download calibration files
        download_calib(root_path)

        # download drives
        for urls in chunks(repos(), threads):
            pool = ThreadPool(threads)
            pool.map(download_unzip, urls)
    except (KeyboardInterrupt, Exception) as e:
        for url in urls:
            # params
            date = date_match.findall(url)[0]
            drive = drive_match.findall(url)[0]

            # clean drive
            drive_path = os.path.join(root_path, date, drive)
            logger.info("removing " + drive_path)
            shutil.rmtree(drive_path, ignore_errors=True)
            if os.path.exists(drive_path + ".zip"):
                os.remove(drive_path + ".zip")
            shutil.rmtree(drive_path + "_tmp", ignore_errors=True)

        # reraise error
        raise e
    finally:
        bar.close()



"""
PyTorch Dataset to load KITTI Raw Data
"""

import glob
import os
import random
import re
from typing import Callable, Dict, Optional, Tuple, Union

from PIL import Image
from torch.utils.data import Dataset
from typing_extensions import Literal

from .calibration import CamCalib, load_imu_to_lidar, load_lidar_to_cam_00
from .inertial_measurement_unit import IMUData
from .lidar_point_cloud import load_lidar_point_cloud
from .synced_rectified import check_drives as sync_rect_check_drives
from .synced_rectified import download as sync_rect_download

__all__ = ["KittiRawDataset"]

_Cams = Union[
    Tuple[str], Tuple[str, str], Tuple[str, str, str], Tuple[str, str, str, str]
]

_Calibs = Union[
    Tuple[str], Tuple[str, str], Tuple[str, str, str], Tuple[str, str, str, str]
]


def _identity(x: Dict) -> Dict:
    return x


class KittiRawDataset(Dataset):
    def __init__(
        self,
        root: str,
        select_cams: _Cams = ("cam_02",),
        imu_data: bool = False,
        lidar_data: Optional[str] = "projective",
        select_calibs: _Calibs = (
            "cam_00",
            "cam_02",
        ),
        load_previous: Union[Tuple[int, int], int] = 0,
        transform: Callable[[Dict], Dict] = _identity,
        download: Union[bool, Literal["sync+rect"]] = False,
    ):
        """
        Dataset loading KITTI Raw Data available, KITTI raw data are organized by date
        and drive and inside each drive are contained images taken by cameras,
        for example::

            .
            └── 2011_09_26
                ├── calib_cam_to_cam.txt
                ├── calib_imu_to_velo.txt
                ├── calib_velo_to_cam.txt
                └── 2011_09_26_drive_0002_sync
                    ├── image_00
                    │   └── data
                    │       └── ...
                    ├── image_01
                    │   └── data
                    │       └── ...
                    ├── image_02
                    │   └── data
                    │       └── ...
                    ├── image_03
                    │   └── data
                    │       └── ...
                    ├── oxts
                    │   └── data
                    │       └── ...
                    └── velodyne_points
                        └── data
                            └── ...

        Calibration data refers to the whole date.
        Each example is a dictionary composed by many entries, such entries
        can be selected at initialization time.

        #. cam_0X: PIL Image from camera X
        #. lidar_data: ndarray Nx4 containing lidar \
                       point cloud with respect to lidar coordinates
        #. lidar_to_cam_00: [R|T] matrix from lidar to cam_00 coordinates in \
                       projective space
        #. cam_0X_calib: :class:`torch_kitti.raw.calibration.CamCalib` object \
                       containing info about cam_0X calibration
        #. imu_data: :class:`torch_kitti.raw.inertial_measurement_unit.IMUData` \
                       object containing IMU data about the example
        #. imu_to_lidar: [R|T] matrix from imu to lidar coordinates in projective \
                       space

        Parameters
        ----------
        root: str
            path to the root of the dataset.
        select_cams: Tuple[str], default ("cam_02",)
            images to be loaded among cam_00, cam_01, cam_02, cam_03.
        imu_data: bool, default False
            if load IMU Data, when true examples will contain "imu_data" and also
            "imu_to_lidar" fields
        lidar_data: str, default "projective"
            if "projective" Lidar Data are loaded in the projective space
            (removing reflectance), if "reflectance" Lidar Data are loaded with
            the reflectance, if None Lidar Data are not loaded. When enabled examples
            will contain "lidar_data" and also "lidar_to_cam_00" fields
        select_calibs: Tuple[str], default ("cam_00, "cam_02")
            calibration objects to be loaded among cam_00, cam_01, cam_02, cam_03, they
            are instances of :class:`torch_kitti.raw_calibration.CamCalib`
        load_previous: Union[int, Tuple[int, int]], optional
            if used a previous nth frame from the same sequence is provided, a random
            previous frame in the range (n, m) is choosen if provided a tuple.
        transform: Callable[[Dict], Dict], optional
            transformation applied to each output dictionary
        download: bool or sync+rect, default False
            if True downloads the sync+rect version, otherwise the version of
            the dataset can be provided but by now only "sync+rect" is available
        """

        # PARAMS

        self.select_cams = set(select_cams)
        self.imu_data = imu_data
        self.transform = transform
        self._load_previous = load_previous
        self.lidar_data = lidar_data
        self.select_calibs = select_calibs

        # CHECK params

        for cam in self.select_cams:
            if cam not in ["cam_00", "cam_01", "cam_02", "cam_03"]:
                raise ValueError("each cam must be in range 0-3")

        for calib in self.select_calibs:
            if calib not in ["cam_00", "cam_01", "cam_02", "cam_03"]:
                raise ValueError("each calib must be in range 0-3")

        if lidar_data not in ["projective", "reflectance", None]:
            raise ValueError("lidar data must be 'projective', 'reflectance' or None")

        if not isinstance(load_previous, (tuple, int)):
            raise ValueError(
                "load_previous must be and integer of a 2-tuple of integers"
            )

        if download not in [False, True, "sync+rect"]:
            raise ValueError("download in True, False or sync+rect")

        # DOWNLOAD

        if download is True or download == "sync+rect":

            # download sync+rect folders
            if not os.path.exists(root):
                sync_rect_download(root)
            elif os.path.isdir(root) and not os.listdir(root):
                sync_rect_download(root)

        # check folders
        if not sync_rect_check_drives(root):
            raise ValueError(f"path {root} contains wrong data")

        # PATHS LOADING

        # load cam images
        cam_00 = list(
            filter(
                lambda f: "image_00" in f,
                glob.iglob(os.path.join(root, "**", "*.png"), recursive=True),
            )
        )
        cam_01 = [f.replace("image_00", "image_01", 1) for f in cam_00]
        cam_02 = [f.replace("image_00", "image_02", 1) for f in cam_00]
        cam_03 = [f.replace("image_00", "image_02", 1) for f in cam_00]

        # load lidar points
        lidar = [
            os.path.splitext(f.replace("image_00", "velodyne_points"))[0] + ".bin"
            for f in cam_00
        ]

        # load oxts
        oxts = [
            os.path.splitext(f.replace("image_00", "oxts"))[0] + ".txt" for f in cam_00
        ]

        # load calib files
        date_match = re.compile("[0-9]+_[0-9]+_[0-9]+")

        def extract_path(f, fname):
            date = date_match.findall(f)[0]
            return os.path.join(root, date, fname)

        cam2cam = [extract_path(f, "calib_cam_to_cam.txt") for f in cam_00]
        velo2cam = [extract_path(f, "calib_velo_to_cam.txt") for f in cam_00]
        imu2velo = [extract_path(f, "calib_imu_to_velo.txt") for f in cam_00]

        self._paths = [
            {
                "cam_00": c00,
                "cam_01": c01,
                "cam_02": c02,
                "cam_03": c03,
                "lidar_point_cloud": l_pc,
                "oxts": o,
                "cam2cam": c2c,
                "velo2cam": v2c,
                "imu2velo": i2v,
            }
            for c00, c01, c02, c03, l_pc, o, c2c, v2c, i2v in zip(
                cam_00, cam_01, cam_02, cam_03, lidar, oxts, cam2cam, velo2cam, imu2velo
            )
        ]

    def __len__(self):
        return len(self._paths)

    def _getitem(self, paths, _can_recur=True):
        output = {}

        # LOAD DATA
        can_load_previous = None

        # load cams
        for cam in self.select_cams:
            can_load_previous = int(os.path.splitext(os.path.basename(paths[cam]))[0])
            output[cam] = Image.open(paths[cam])

        # load config files
        for calib in self.select_calibs:
            idx = int(re.compile("[0-9]{2}").findall(calib)[0])
            calib_file = CamCalib.open(idx, paths["cam2cam"])
            output[f"cam_0{idx}_calib"] = calib_file

        # load lidar points
        if self.lidar_data is not None:
            can_load_previous = int(
                os.path.splitext(os.path.basename(paths["lidar_point_cloud"]))[0]
            )
            output["lidar_data"] = load_lidar_point_cloud(
                paths["lidar_point_cloud"], self.lidar_data == "projective"
            )
            output["lidar_to_cam_00"] = load_lidar_to_cam_00(paths["velo2cam"])

        # IMU Data
        if self.imu_data:
            can_load_previous = int(
                os.path.splitext(os.path.basename(paths["oxts"]))[0]
            )
            output["imu_data"] = IMUData.open(paths["oxts"])
            output["imu_to_lidar"] = load_imu_to_lidar(paths["imu2velo"])

        # LOAD PREVIOUS
        if can_load_previous is not None:
            curr_idx = can_load_previous
            if isinstance(self._load_previous, int) and self._load_previous > 0:
                previous = max(0, curr_idx - self._load_previous)
            elif isinstance(self._load_previous, tuple):
                previous = max(0, curr_idx - random.randint(*self._load_previous))
            else:
                previous = None

            if previous is not None and _can_recur:
                idx = str(previous).zfill(10)
                new_paths = dict()
                for k in paths:
                    new_paths[k] = re.sub(r"[0-9]{10}", idx, paths[k])
                previous = self._getitem(new_paths, _can_recur=False)
                for k in previous:
                    output[k + "_previous"] = previous[k]

        return output

    def __getitem__(self, x):
        return self.transform(self._getitem(self._paths[x]))



"""
Utilities to handle calibration files
"""

from typing import Dict, Tuple

import numpy as np

__all__ = ["CamCalib", "load_imu_to_lidar", "load_lidar_to_cam_00"]

# LOADING CALIBRATION FILES


# cam_to_cam.txt


class CamCalib:
    def __init__(
        self,
        cam: int,
        image_size: Tuple[int, int],
        intrinsics: np.ndarray,
        distortion: np.ndarray,
        extrinsics: np.ndarray,
        rect_image_size: Tuple[int, int],
        rect_rotation: np.ndarray,
        projection_matrix: np.ndarray,
    ):
        """
        Calibration metrics for a single camera.

        Parameters
        ----------
        cam: int
            the choosen camera among 0, 1, 2, 3
        image_size: (int, int)
            size of the image before rectification
        intrinsics: ndarray
            3x3 array containing the intrinsics parameters
            of the camera
        distortion: ndarray
            k1, k2, p1, p2, k3 distortion coefficients where k1, k2 and k3 are
            the radial coeeficients. p1 and p2 are the tangential distortion
            coefficients
        extrinsics: ndarray
            4x4 array containing rototraslation matrix in the projective space,
            They seem to be a transformation from a common worl coordinate system
            into the camera's coordinate system
        rect_image_size: (int, int)
            size of the image after rectification
        rect_rotation: ndarray
            3x3 rotation matrix performing rectigying rotation for reference coordinate
            to make images of multiple cameras lie on the same plan
        projection_matrix: ndarray
            the projection matrix from 3D to 2D after rectification.
        """
        self.cam = cam
        self.image_size = image_size
        self.intrinsics = intrinsics
        self.distortion = distortion
        self.extrinsics = extrinsics
        self.rect_image_size = rect_image_size
        self.rect_rotation = rect_rotation
        self.projection_matrix = projection_matrix

    @staticmethod
    def _to_homologous_coord(rot_matrix):
        R = np.eye(4)
        R[:3, :3] = rot_matrix
        return R

    @staticmethod
    def open(cam: int, path: str) -> "CamCalib":

        # load data
        dict_values: Dict[str, np.ndarray] = dict()
        with open(path, "rt") as f:
            for line in f.readlines():
                name, vector = line.split(":", maxsplit=1)
                if name == "calib_time":
                    continue
                vector = np.array(list(map(float, vector.strip().split(" "))))
                dict_values[name] = vector

        # convert
        cams = "0" + str(cam)

        image_size = dict_values[f"S_{cams}"].astype(np.int)
        image_size = (int(image_size[0]), int(image_size[1]))

        rect_image_size = dict_values[f"S_rect_{cams}"].astype(np.int)
        rect_image_size = (int(rect_image_size[0]), int(rect_image_size[1]))

        R = dict_values[f"R_{cams}"].reshape(3, 3)
        T = dict_values[f"T_{cams}"]
        extrinsics = np.eye(4)
        extrinsics[:3, :3] = R
        extrinsics[:3, 3] = T

        return CamCalib(
            cam=cam,
            image_size=image_size,
            intrinsics=dict_values[f"K_{cams}"].reshape(3, 3),
            distortion=dict_values[f"D_{cams}"],
            extrinsics=extrinsics,
            rect_image_size=rect_image_size,
            rect_rotation=CamCalib._to_homologous_coord(
                dict_values[f"R_rect_{cams}"].reshape(3, 3)
            ),
            projection_matrix=dict_values[f"P_rect_{cams}"].reshape(3, 4),
        )


# imu_to_velo.txt & calib_velo_to_cam.txt


def _load_rototraslation(path: str) -> np.ndarray:
    dict_values: Dict[str, np.ndarray] = dict()
    with open(path, "rt") as f:
        for line in f.readlines():
            name, vector = line.split(":", maxsplit=1)
            if name == "calib_time":
                continue
            vector = np.array(list(map(float, vector.strip().split(" "))))
            dict_values[name] = vector

    R = dict_values["R"].reshape(3, 3)
    T = dict_values["T"]
    rt = np.eye(4)
    rt[:3, :3] = R
    rt[:3, 3] = T

    return rt


def load_imu_to_lidar(path: str) -> np.ndarray:
    """
    returns the rototraslation matrix in projective coordinates from the
    inertial measurement unit to the lidar position. Done to read
    imu_to_velo.txt files.
    """
    return _load_rototraslation(path)


def load_lidar_to_cam_00(path: str) -> np.ndarray:
    """
    returns the rototraslation matrix in projective coordinates from the
    lidar to camera 00. Done to read velo_to_cam.txt files.
    """
    return _load_rototraslation(path)



"""
Utilities to handle the KITTI Vision Benchmark Suite in PyTorch
"""

__version__ = "0.2.3"

__all__ = ["raw", "depth_completion", "depth_prediction", "transforms", "metrics"]

from . import depth_completion, depth_prediction, metrics, raw, transforms



from typing import Any, Callable, Dict, List, Optional

from .functional import add_features, apply_to_features

__all__ = ["ApplyToFeatures", "AddFeatures"]


class ApplyToFeatures:
    """
    Apply the same transformation to a subset of features of a dictionary.
    Even random transformations can be applied in the same way to each
    field cause the inner random state is the same for each application.

    Parameters
    ----------
    features: List[Any], optional
        A list of dictionary keys to apply `transform`, if not provided `transform`
        is applied to all fields
    transform: Callable
        Transformation to be applied
    same_rand_state: bool, default True
        If use the same random state for each application of `transform`.

    Returns
    -------
    The same input with specified `features` transformated by `transform`

    Example
    -------

    >>> d = {"a": np.random.randn(1, 3, 100, 100), "b": np.random.randn(1, 3, 100, 100)}
    >>> d = ApplyToFeatures(lambda x: x + np.random.rand(), ["a", "b"])(d)

    """

    def __init__(
        self,
        transform: Callable[[Any], Any],
        features: Optional[List[Any]] = None,
        same_rand_state: bool = True,
    ):
        self.features = features
        self.transform = transform
        self.same_rand_state = same_rand_state

    def __call__(self, x: Dict) -> Dict:
        return apply_to_features(self.transform, x, self.features, self.same_rand_state)


class AddFeatures:
    """
    Takes the input, passes it to a transformation and merge the original
    input with the result of the transformation, can be used to
    augment data in the batch.

    .. note::
        transform type and input type must match.

    Parameters
    ----------
    transform: Callable
        transformation applied
    x: Dict
        data to transform

    Example
    -------

    >>> add_features(lambda _: {'y': 1}, {'a': 2})
    {'a': 2, 'y': 1}
    """

    def __init__(self, transform: Callable[[Dict], Dict]):
        self.transform = transform

    def __call__(self, x: Dict) -> Dict:
        return add_features(self.transform, x)



from typing import Union

import torch

# DEPTH COMPLETION METRICS

__all__ = [
    "rmse",
    "mse",
    "irmse",
    "mae",
    "imae",
    "sq_rel_error",
    "abs_rel_error",
    "silog",
    "ratio_threshold",
]


def rmse(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    """
    computes the root mean squared error
    """
    return torch.sqrt(torch.mean(torch.square(y_pred - y_true)))


def mse(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    """
    computes the mean squared error
    """
    return torch.mean(torch.square(y_pred - y_true))


def irmse(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    r"""
    Compute the inverse root mean squared error

    .. math::

        \sqrt{\frac{1}{|V|}\sum_{v \in V}|\frac{1}{d_v^{gt}} - \frac{1}{d_v^{pred}}|^2}
    """
    return torch.sqrt(torch.mean(torch.square(1 / y_true - 1 / y_pred)))


def mae(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    """
    computes the mean absolute error
    """
    return torch.mean(torch.abs(y_pred - y_true))


def imae(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    r"""
    Compute inverse mean absolute error

    .. math::

        \frac{1}{|V|}\sum_{v \in V}|\frac{1}{d_v^{gt}} - \frac{1}{d_v^{pred}}|
    """
    return torch.mean(torch.abs(1 / y_true - 1 / y_pred))


# DEPTH PREDICTION METRICS


def sq_rel_error(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    r"""
    Computes the relative L2 error with respect to
    y_true

    .. math::

        \frac{(y - \hat{y})^2}{\hat{y}}
    """
    return torch.mean(torch.square(y_pred - y_true) / y_true)


def abs_rel_error(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    r"""
    Computes the relative L1 error with respect to
    y_true

    .. math::

        \frac{|y - \hat{y}|}{\hat{y}}
    """
    return torch.mean(torch.abs(y_pred - y_true) / y_true)


def silog(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    r"""
    Computes the scale invariant logarithmig error as states here_

    .. _here: https://arxiv.org/abs/1406.2283
    """

    return torch.mean(
        torch.square(torch.log(y_true) - torch.log(y_pred))
    ) - torch.square(torch.mean(torch.log(y_true) - torch.log(y_pred)))


def ratio_threshold(
    threshold: Union[torch.Tensor, float], y_pred: torch.Tensor, y_true: torch.Tensor
) -> torch.Tensor:
    r"""
    Computes the percentage of values whose maximum between thre ratio and the inverse
    ratio with respect to the `y_true` is lower than the `threshold`.
    """
    result = torch.max(y_true / y_pred, y_pred / y_true)
    result = (result < threshold).to(torch.float32).mean()
    return result



import random
from typing import Callable, Dict, List, Optional

import numpy as np
import torch

__all__ = ["apply_to_features", "add_features"]


def apply_to_features(
    transform: Callable,
    x: Dict,
    features: Optional[List[str]] = None,
    same_rand_state=True,
) -> Dict:
    """
    Apply the same transformation to a subset of features of a dictionary.
    Even random transformations can be applied in the same way to each
    field cause the inner random state is the same for each application.

    Parameters
    ----------
    features: List[Any], optional
        A list of dictionary keys to apply `transform`, if not provided `transform`
        is applied to all fields
    transform: Callable
        Transformation to be applied
    same_rand_state: bool, default True
        If use the same random state for each application of `transform`.

    Returns
    -------
    The same input with specified `features` transformated by `transform`

    Example
    -------

    >>> d = {"a": np.random.randn(1, 3, 100, 100), "b": np.random.randn(1, 3, 100, 100)}
    >>> d = apply_to_features(lambda x: x + np.random.rand(), d, ["a", "b"])

    """

    if same_rand_state:

        # move randomness
        np.random.rand()
        random.random()
        torch.rand(1)

        # save state
        np_state = np.random.get_state()
        rd_state = random.getstate()
        tr_state = torch.random.get_rng_state()

    y = dict(**x)

    if features is None:
        features = list(x.keys())

    for feature in features:
        y[feature] = transform(x[feature])

        if same_rand_state:
            np.random.set_state(np_state)
            random.setstate(rd_state)
            torch.set_rng_state(tr_state)

    return y


def add_features(transform: Callable[[Dict], Dict], x: Dict) -> Dict:
    """
    Takes the input, passes it to a transformation and merge the original
    input with the result of the transformation, can be used to
    augment data in the batch.

    .. note::
        transform type and input type must match.

    Parameters
    ----------
    transform: Callable
        transformation applied
    x: Dict
        data to transform

    Example
    -------

    >>> add_features(lambda _: {'y': 1}, {'a': 2})
    {'a': 2, 'y': 1}
    """
    results = transform(x)
    if isinstance(results, dict) and isinstance(x, dict):
        return dict(x, **results)
    else:
        raise ValueError("dictionaries supported only")



"""
KITTI Vision Benchmark Raw Dataset Manipulation
"""

__all__ = [
    "CamCalib",
    "IMUData",
    "synced_rectified",
    "KittiRawDataset",
    "load_lidar_point_cloud",
]

from . import synced_rectified
from .calibration import CamCalib
from .dataset import KittiRawDataset
from .inertial_measurement_unit import IMUData
from .lidar_point_cloud import load_lidar_point_cloud



"""
Download utilities
"""

__all__ = ["calib_repos"]

import os
from pathlib import Path
from typing import List
from zipfile import ZipFile

import requests
from tqdm import tqdm


def calib_repos() -> List[str]:
    """
    Provides a plain list of urls pointing to all calib files of KITTI dataset.
    These are also available at the address here_.

    .. _here: http://www.cvlibs.net/datasets/kitti/raw_data.php
    """

    repos = [
        "https://s3.eu-central-1.amazonaws.com/avg-kitti/raw_data/2011_09_30_calib.zip",
        "https://s3.eu-central-1.amazonaws.com/avg-kitti/raw_data/2011_09_29_calib.zip",
        "https://s3.eu-central-1.amazonaws.com/avg-kitti/raw_data/2011_09_26_calib.zip",
        "https://s3.eu-central-1.amazonaws.com/avg-kitti/raw_data/2011_10_03_calib.zip",
        "https://s3.eu-central-1.amazonaws.com/avg-kitti/raw_data/2011_09_28_calib.zip",
    ]

    return repos


def download_file(url, save_path, chunk_size=1024, verbose=True):
    """
    Downloads a zip file from an `url` into a zip file in the
    provided `save_path`.
    """
    r = requests.get(url, stream=True)
    zip_name = url.split("/")[-1]

    content_length = int(r.headers["Content-Length"]) / 10 ** 6

    if verbose:
        bar = tqdm(total=content_length, unit="Mb", desc="download " + zip_name)
    with open(save_path, "wb") as fd:
        for chunk in r.iter_content(chunk_size=chunk_size):
            fd.write(chunk)
            if verbose:
                bar.update(chunk_size / 10 ** 6)

    if verbose:
        bar.close()


def download_calib(root_path: str):
    """
    Downloads and scaffolds calibration files
    """

    Path(root_path).mkdir(exist_ok=True)

    for repo in calib_repos():
        calib_zip_path = os.path.join(root_path, "calib.zip")
        download_file(repo, calib_zip_path)
        with open(calib_zip_path, "rb") as f:
            ZipFile(f).extractall(root_path)
        os.remove(calib_zip_path)



"""
testing integration between torchvision and transformations
"""

import torch
import torchvision.transforms as V

import torch_kitti.transforms as K


def test_random_crop():

    # simulate input
    fake_img_1 = torch.randn(1, 600, 600)
    fake_img_2 = fake_img_1.clone()
    x = {"img_left": fake_img_1, "img_right": fake_img_2}

    output = K.functional.apply_to_features(V.RandomCrop([200, 200]), x)
    assert torch.all(output["img_left"] == output["img_right"])



import logging
import os
from zipfile import ZipFile

import requests
from tqdm import tqdm

__all__ = ["download", "folders_check"]

logger = logging.getLogger(__name__)


def download_file(url, save_path, chunk_size=1024, verbose=True):
    """
    Downloads a zip file from an `url` into a zip file in the
    provided `save_path`.
    """
    r = requests.get(url, stream=True)
    zip_name = url.split("/")[-1]

    content_length = int(r.headers["Content-Length"]) / 10 ** 6

    if verbose:
        bar = tqdm(total=content_length, unit="Mb", desc="download " + zip_name)

    try:
        with open(save_path, "wb") as fd:
            for chunk in r.iter_content(chunk_size=chunk_size):
                fd.write(chunk)
                if verbose:
                    bar.update(chunk_size / 10 ** 6)
    except (Exception, KeyboardInterrupt) as e:
        os.remove(save_path)
        raise e
    finally:
        if verbose:
            bar.close()


def download(root_path: str, progress_bar=True):
    """
    Downloads and scaffold depth completion dataset in `root_path`
    """

    if not os.path.exists(root_path):
        os.mkdir(root_path)

    # urls
    data_depth_selection_url = (
        "https://s3.eu-central-1.amazonaws.com/avg-kitti/data_depth_selection.zip"
    )
    data_depth_velodyne_url = (
        "https://s3.eu-central-1.amazonaws.com/avg-kitti/data_depth_velodyne.zip"
    )
    data_depth_annotated_url = (
        "https://s3.eu-central-1.amazonaws.com/avg-kitti/data_depth_annotated.zip"
    )

    # download of zips
    download_file(
        data_depth_selection_url,
        os.path.join(root_path, "data_depth_selection.zip"),
        verbose=progress_bar,
    )

    download_file(
        data_depth_velodyne_url,
        os.path.join(root_path, "data_depth_velodyne.zip"),
        verbose=progress_bar,
    )
    download_file(
        data_depth_annotated_url,
        os.path.join(root_path, "data_depth_annotated.zip"),
        verbose=progress_bar,
    )

    # unzip and remove zips
    logging.info("unzipping...")

    with ZipFile(os.path.join(root_path, "data_depth_selection.zip"), "r") as zip_ref:
        zip_ref.extractall(root_path)
        os.rename(
            os.path.join(
                root_path, "depth_selection", "test_depth_completion_anonymous"
            ),
            os.path.join(root_path, "test_depth_completion_anonymous"),
        )
        os.rename(
            os.path.join(
                root_path, "depth_selection", "test_depth_prediction_anonymous"
            ),
            os.path.join(root_path, "test_depth_prediction_anonymous"),
        )
        os.rename(
            os.path.join(root_path, "depth_selection", "val_selection_cropped"),
            os.path.join(root_path, "val_selection_cropped"),
        )
        os.rmdir(os.path.join(root_path, "depth_selection"))
    with ZipFile(os.path.join(root_path, "data_depth_velodyne.zip"), "r") as zip_ref:
        zip_ref.extractall(root_path)
    with ZipFile(os.path.join(root_path, "data_depth_annotated.zip"), "r") as zip_ref:
        zip_ref.extractall(root_path)

    # remove zip files
    os.remove(os.path.join(root_path, "data_depth_selection.zip"))
    os.remove(os.path.join(root_path, "data_depth_completion.zip"))
    os.remove(os.path.join(root_path, "data_depth_velodyne.zip"))

    logging.info("done.")


def folders_check(root_path: str) -> bool:
    """
    Performs some simple checks about folders structure for depth completion dataset and
    prints errors
    """

    for folder in [
        "test_depth_completion_anonymous",
        "test_depth_prediction_anonymous",
        "val_selection_cropped",
    ]:
        if not os.path.exists(os.path.join(root_path, folder)):
            logging.error(
                f"missing data_depth_selection.zip: folder {folder} not found"
            )
            return False

    # check of data_depth_completion.zip and data_depth_velodyne.zip
    def check_gt_velodyne(folder):
        num_groundtruth, num_velodyne = 0, 0
        for dirpath, dirnames, filenames in os.walk(os.path.join(root_path, folder)):
            if "groundtruth" in dirnames:
                num_groundtruth += 1
            if "velodyne_raw" in dirnames:
                num_velodyne += 1
        return num_groundtruth, num_velodyne

    gt_train, velodyne_train = check_gt_velodyne("train")
    if gt_train != 138 and velodyne_train != 138:
        logger.error(
            "something went wrong with data_depth_completion.zip and \
             data_depth_velodyne.zip in val folder"
        )

        logger.error(
            f"found {gt_train} groundtruth folders and \
             {velodyne_train} raw_lidar folders but they should be 138.."
        )

        return False

    gt_val, velodyne_val = check_gt_velodyne("val")
    if gt_val != 13 and velodyne_val != 13:
        logger.error(
            "something went wrong with data_depth_completion.zip and \
            data_depth_velodyne.zip in val folder"
        )

        logger.error(
            f"found {gt_train} groundtruth folders and {velodyne_train} \
             raw_lidar folders but the should be 13.."
        )

        return False

    return True



"""
Dataset loading for Depth Completion KITTI Dataset
:ref:`http://www.cvlibs.net/datasets/kitti/eval_depth.php?benchmark=depth_completion`
"""

import errno
import os
import random
import re
from typing import Callable, Dict, Tuple, Union

import numpy as np
from PIL import Image
from torch.utils.data import Dataset
from typing_extensions import Literal

from torch_kitti.depth_completion import download as kitti_depth_completion_download
from torch_kitti.depth_completion import (
    folders_check as kitti_depth_completion_folders_check,
)
from torch_kitti.raw.calibration import CamCalib
from torch_kitti.raw.synced_rectified import check_drives as kitti_raw_check_drives
from torch_kitti.raw.synced_rectified import download as kitti_raw_download

__all__ = ["KittiDepthCompletionDataset"]

_Cams = Union[
    Tuple[str], Tuple[str, str], Tuple[str, str, str], Tuple[str, str, str, str]
]

_Calibs = Union[
    Tuple[str], Tuple[str, str], Tuple[str, str, str], Tuple[str, str, str, str]
]


def _identity(x: Dict) -> Dict:
    return x


class KittiDepthCompletionDataset(Dataset):
    """
    2017 KITTI depth completion benchmarks
    dataset, consisting of 93k training and 1.5k test images.

    Ground truth has been acquired by accumulating 3D point clouds from a
    360 degree Velodyne HDL-64 Laserscanner and a consistency check using
    stereo camera pairs.

    To load this dataset are needed:

    KITTI Raw dataset (sync+rect)
        can be found at :ref:`http://www.cvlibs.net/datasets/kitti/raw_data.php`,
        automatically downloaded if required.

    KITTI depth completion maps
        can be found at :ref:`http://www.cvlibs.net/datasets/kitti/eval_depth.php?\
        benchmark=depth_completion`, automatically downloaded if required

    Parameters
    ----------
    kitti_raw_root: str
        path to the root of the KITTI raw data sync+rect folder.
    depth_completion_root: str
        path to the root of the depth completion path.
    subset: str, default train
        If 'train' creates the dataset from training set, if 'val' creates
        the dataset from validation set, if 'test' creates the
        dataset from test set.
    load_stereo: bool, default False
        if True each batch provides left and right synchronized and rectified
        cameras (the dataset size is halved). Not available on testing.
    load_previous: Union[int, Tuple[int, int]], optional
        if used a previous nth frame from the same sequence is provided, a random
        previous frame in the range (n, m) is choosen if provided a tuple. if such
        frame cannot be found the same frame is returned as previous
    transform: Callable[[Dict], Dict], optional
        transformation applied to each output dictionary.
    download: bool, default False
        If true, downloads the dataset from the internet and puts
        it in root directories. If dataset is already downloaded,
        it is not downloaded again.

    .. note::
        On the test subset `load_stereo` and `load_previous` can not be used
    """

    def __init__(
        self,
        kitti_raw_root: str,
        depth_completion_root: str,
        subset: Literal["train", "val", "test"] = "train",
        load_stereo: bool = False,
        load_previous: Union[Tuple[int, int], int] = 0,
        transform: Callable[[Dict], Dict] = _identity,
        download: bool = False,
    ):

        # PARAMS

        self._subset = subset
        self._load_stereo = load_stereo
        self._load_previous = load_previous
        self.transform = transform

        if subset not in ["train", "val", "test"]:
            raise ValueError("subset must be in train, val, test")
        if not isinstance(load_previous, int) and not isinstance(load_previous, tuple):
            raise ValueError("load_previous int or tuple")
        if isinstance(load_previous, int) and load_previous < 0:
            raise ValueError("load_previous >= 0")

        # DOWNLOAD

        if download:
            # download depth completion dataset
            if not os.path.exists(depth_completion_root):
                kitti_depth_completion_download(depth_completion_root)
            elif os.path.isdir(depth_completion_root) and not os.listdir(
                depth_completion_root
            ):
                kitti_depth_completion_download(depth_completion_root)

            # download kitti raw rect+sync dataset
            if not os.path.exists(kitti_raw_root):
                kitti_raw_download(kitti_raw_root)
            elif os.path.isdir(kitti_raw_root) and not os.listdir(kitti_raw_root):
                kitti_raw_download(kitti_raw_root)

        # check folders
        if not kitti_depth_completion_folders_check(depth_completion_root):
            raise ValueError(f"path {depth_completion_root} contains wrong data")
        if not kitti_raw_check_drives(kitti_raw_root):
            raise ValueError(f"path {kitti_raw_root} contains wrong data")

        # computing paths
        if subset in ["train", "val"]:

            depth_completion_path = os.path.join(depth_completion_root, subset)

            # find lidar and groundtruths
            all_files = []
            for dirpath, dirnames, filenames in os.walk(depth_completion_path):
                files = [os.path.join(dirpath, f) for f in filenames]
                all_files.extend(files)

            all_lidar_raws = [f for f in all_files if "velodyne_raw" in f]
            groundtruths = [f for f in all_files if "groundtruth" in f]

            # build matching image paths
            drive_regex = re.compile(r"[0-9]+_[0-9]+_[0-9]+_drive_[0-9]+_sync")
            date_regex = re.compile(r"[0-9]+_[0-9]+_[0-9]+")
            camera_regex = re.compile(r"image_[0-9]+")

            images = []
            for groundtruth_path in groundtruths:
                try:
                    drive = drive_regex.findall(groundtruth_path)[0]
                    date = date_regex.findall(groundtruth_path)[0]
                    camera = camera_regex.findall(groundtruth_path)[0]
                    image = groundtruth_path.split(os.path.sep)[-1]
                    images.append(
                        os.path.join(kitti_raw_root, date, drive, camera, "data", image)
                    )
                except IndexError:
                    raise RuntimeError("wrong path {}".format(groundtruth_path))

            # compute intrinsics
            cam2cam = []
            for img in images:
                date = date_regex.findall(img)[0]
                cam2cam.append(
                    os.path.join(kitti_raw_root, date, "calib_cam_to_cam.txt")
                )

            self._paths = [
                {"img": img, "gt": gt, "lidar": lidar, "cam2cam": c2c}
                for img, gt, lidar, c2c in zip(
                    images, groundtruths, all_lidar_raws, cam2cam
                )
            ]

        elif subset == "test":

            folders = os.path.join(
                depth_completion_root, "val_selection_cropped", "{0}"
            )
            files = [
                os.path.join(folders, f.replace("groundtruth_depth", "{1}"))
                for f in os.listdir(folders.format("groundtruth_depth"))
                if os.path.isfile(os.path.join(folders.format("groundtruth_depth"), f))
            ]

            self._paths = [
                {
                    "img": path.format("image", "image"),
                    "gt": path.format("groundtruth_depth", "groundtruth_depth"),
                    "lidar": path.format("velodyne_raw", "velodyne_raw"),
                    "intrinsics": os.path.splitext(path.format("intrinsics", "image"))[
                        0
                    ]
                    + ".txt",
                }
                for path in files
            ]

        else:
            raise ValueError("subset must be train, val or test. Not {}".format(subset))

        if load_stereo and subset == "test":
            raise ValueError("on test data stereo is not available")
        elif load_stereo:
            for path in self._paths:
                for key in path:
                    path[key] = re.sub("image_0[23]", "{0}", path[key])
            self._paths = list({v["img"]: v for v in self._paths}.values())

        if (
            isinstance(load_previous, (tuple))
            or isinstance(load_previous, int)
            and load_previous > 0
        ) and subset == "test":
            raise ValueError("on test data previous frame is not available")

    def _open_img(self, path):
        extensions = [".png", ".jpg"]

        image = None
        for ext in extensions:
            try:
                image = Image.open(os.path.splitext(path)[0] + ext)
            except FileNotFoundError:
                pass

        if image is None:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

        return image

    def _getitem(self, paths, _can_recur=True):
        result = dict()

        cam_used = int(re.match(".*image_([0-9]+).*", paths["img"]).groups()[0])

        # open image
        img = self._open_img(paths["img"])
        result["img"] = img

        # compute lidar points
        if "lidar" in paths:
            lidar = self._open_img(paths["lidar"])
            lidar = np.array(lidar).astype(np.float32) / 256.0
            lidar = np.expand_dims(lidar, axis=-1)
            result["lidar"] = lidar

        # compute groundtruth points
        gt = self._open_img(paths["gt"])
        gt = np.array(gt).astype(np.float32) / 256.0
        gt = np.expand_dims(gt, axis=-1)
        result["gt"] = gt

        # compute intrinsics
        if "cam2cam" in paths:
            calib = CamCalib.open(cam_used, paths["cam2cam"])
            result["intrinsics"] = calib.projection_matrix[:, :-1]
        elif "intrinsics" in paths:
            with open(paths["intrinsics"], "rt") as f:
                result["intrinsics"] = np.array(
                    list(map(float, f.readline().strip().split(" ")))
                ).reshape(3, 3)

        # compute previous
        if isinstance(self._load_previous, int) and self._load_previous > 0:
            curr_img_idx = int(
                re.match(r".*([0-9]{10}).png$", paths["img"]).groups()[0]
            )
            previous = max(0, curr_img_idx - self._load_previous)
            if not os.path.exists(
                re.sub(r"\d{10}", str(previous).zfill(10), paths["gt"])
            ):
                previous = curr_img_idx

        elif isinstance(self._load_previous, tuple):
            curr_img_idx = int(
                re.match(r".*([0-9]{10}).png$", paths["img"]).groups()[0]
            )
            previous = max(0, curr_img_idx - random.randint(*self._load_previous))
            if not os.path.exists(
                re.sub(r"\d{10}", str(previous).zfill(10), paths["gt"])
            ):
                previous = curr_img_idx

        else:
            previous = None

        if previous is not None and _can_recur:
            idx = str(previous).zfill(10)
            new_paths = dict()
            for k in paths:
                new_paths[k] = re.sub(r"[0-9]{10}", idx, paths[k])
            previous = self._getitem(new_paths, _can_recur=False)
            for k in previous:
                result[k + "_previous"] = previous[k]

        return result

    def __getitem__(self, x):

        if self._load_stereo:
            # load left
            left_paths = dict(**self._paths[x])
            for k in left_paths:
                left_paths[k] = left_paths[k].format("image_02")
            left_ex = self._getitem(left_paths)

            # load right
            right_paths = dict(**self._paths[x])
            for k in right_paths:
                right_paths[k] = right_paths[k].format("image_03")
            right_ex = self._getitem(right_paths)

            # rename and return
            left_ex_, right_ex_ = dict(), dict()
            for k_l, k_r in zip(left_ex.keys(), right_ex.keys()):
                left_ex_[k_l + "_left"] = left_ex[k_l]
                right_ex_[k_r + "_right"] = right_ex[k_r]

            return self.transform(dict(**left_ex_, **right_ex_))
        else:
            return self.transform(self._getitem(self._paths[x]))

    def __len__(self):
        return len(self._paths)



"""
Utilities to load IMU Data
"""

import numpy as np

__all__ = ["IMUData"]


class IMUData:
    def __init__(self, **kwargs):
        for kw, v in kwargs.items():
            setattr(self, kw, v)

    @staticmethod
    def open(path: str) -> "IMUData":
        fields = [
            "lat",
            "lon",
            "alt",
            "roll",
            "pitch",
            "yaw",
            "vn",
            "ve",
            "vf",
            "vl",
            "vu",
            "ax",
            "ay",
            "ay",
            "af",
            "al",
            "au",
            "wx",
            "wy",
            "wz",
            "wf",
            "wl",
            "wu",
            "pos_accuracy",
            "vel_accuracy",
            "navstat",
            "numsats",
            "posmode",
            "velmode",
            "orimode",
        ]
        with open(path, "rt") as f:
            vector = f.readline()
            vector = np.array(list(map(float, vector.strip().split(" "))))

        return IMUData(**dict(zip(fields, vector)))



"""
Utilities to load and manipulate Lidar Point clouds
"""

import numpy as np

__all__ = ["load_lidar_point_cloud"]


def load_lidar_point_cloud(path: str, projective: bool = True) -> np.ndarray:
    """
    Loads points from the binary file generated
    by a lidar and returns xyz projective coordinates

    Parameters
    ----------
    path: str
        path to the file containing lidar points, in KITTI raw dataset
        these files are under 'velodyne_points/data' folder with .bin
        extension.
    projective: bool, default True
        lidar points also contain also reflectance for each point, if
        `projective` is True reflectance is substituted by '1.0' in
        order to have projective coordinates.

    Returns
    -------
    ndarray
        returns a numpy ndarray of shape [N, 4] of N points with x y z 1
        coordinates if projective is True otherwise 1 is substituted by
        reflectance

    Examples
    --------

    >>> load_lidar_points('0000000000.bin').shape
    (60540, 4)

    """
    points = np.fromfile(path, dtype=np.float32).reshape(-1, 4)

    # last value should be reflectance. Is changed in projective coordinates
    if projective:
        points[:, 3] = 1.0
    return points



"""
Dataset loading for Depth Completion KITTI Dataset
:ref:`http://www.cvlibs.net/datasets/kitti/eval_depth.php?benchmark=depth_completion`
"""

__all__ = ["KittiDepthPredictionDataset", "download", "folders_check"]

from ..depth_completion import download, folders_check
from .dataset import KittiDepthPredictionDataset



"""
Main program for torch kitti
"""

from argparse import ArgumentParser

from torch_kitti.depth_completion import download as depth_completion_download
from torch_kitti.depth_prediction import download as depth_prediction_download
from torch_kitti.raw.synced_rectified import download as sync_rect_download

# dowload datasets

download_funcs = {
    "sync_rectified": sync_rect_download,
    "depth_completion": depth_completion_download,
    "depth_prediction": depth_prediction_download,
}


def main():

    # parser
    parser = ArgumentParser("Torch Kitti")
    subparsers = parser.add_subparsers(help="sub-command help")

    # download subparser
    download_parser = subparsers.add_parser("download", help="scaffolding utilities")
    download_parser.add_argument(
        "name",
        type=str,
        choices=download_funcs.keys(),
        help="name of the dataset to download",
    )
    download_parser.add_argument("path", type=str, help="where scaffold the dataset")
    download_parser.set_defaults(func=lambda args: download_funcs[args.name](args.path))

    # run
    args = parser.parse_args()
    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
import re

from setuptools import find_packages, setup

with io.open("torch_kitti/__init__.py", "rt", encoding="utf8") as f:
    searched = re.search(r"__version__ = \"(.*?)\"", f.read())
    if searched is None:
        raise ImportError("Could not find __version__ in torch_kitti/__init__.py")
    else:
        version = searched.group(1)

setup(
    name="torch-kitti",
    version=version,
    author="Andrea Conti",
    author_email="andrea.conti@tutanota.com",
    description="utilities and pytorch datasets for the KITTI Vision Benchmark Suite",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(),
    entry_points={
        "console_scripts": ["torch_kitti = torch_kitti.scripts.torch_kitti:main"]
    },
    package_data={"": ["*.txt", "*.pkl"]},
    install_requires=[
        "torch >= 1.5.0",
        "torchvision >= 0.6.0",
        "setuptools >= 46.3.0",
        "Pillow > 7.1.0",
        "requests >= 2.24.0",
        "tqdm >= 4.51.0",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pytest-cov",
            "pre-commit",
            "mypy",
            "flake8",
            "black",
            "isort",
            "tox",
        ],
        "doc": ["sphinx", "sphinx-rtd-theme"],
    },
)



"""
Dataset loading for Depth Completion KITTI Dataset
:ref:`http://www.cvlibs.net/datasets/kitti/eval_depth.php?benchmark=depth_prediction`
"""

from typing import Callable, Dict, Tuple, Union

from typing_extensions import Literal

from ..depth_completion.dataset import KittiDepthCompletionDataset

_Cams = Union[
    Tuple[str], Tuple[str, str], Tuple[str, str, str], Tuple[str, str, str, str]
]

_Calibs = Union[
    Tuple[str], Tuple[str, str], Tuple[str, str, str], Tuple[str, str, str, str]
]


class KittiDepthPredictionDataset(KittiDepthCompletionDataset):
    """
    2017 KITTI depth prediction benchmarks
    dataset, consisting of 93k training and 1.5k test images.

    Ground truth has been acquired by accumulating 3D point clouds from a
    360 degree Velodyne HDL-64 Laserscanner and a consistency check using
    stereo camera pairs.

    To load this dataset are needed:

    KITTI Raw dataset (sync+rect)
        can be found at :ref:`http://www.cvlibs.net/datasets/kitti/raw_data.php`,
        automatically downloaded if required.

    KITTI depth prediction maps
        can be found at :ref:`http://www.cvlibs.net/datasets/kitti/eval_depth.php?\
        benchmark=depth_prediction`, automatically downloaded if required

    Parameters
    ----------
    kitti_raw_root: str
        path to the root of the KITTI raw data sync+rect folder.
    depth_prediction_root: str
        path to the root of the depth prediction path.
    subset: str, default train
        If 'train' creates the dataset from training set, if 'val' creates
        the dataset from validation set, if 'test' creates the
        dataset from test set.
    load_stereo: bool, default False
        if True each batch provides left and right synchronized and rectified
        cameras (the dataset size is halved). Not available on testing.
    load_previous: Union[int, Tuple[int, int]], optional
        if used a previous nth frame from the same sequence is provided, a random
        previous frame in the range (n, m) is choosen if provided a tuple. If that
        frame is not available is provided the same frame twice.
    transform: Callable[[Dict], Dict], optional
        transformation applied to each output dictionary.
    download: bool, default False
        If true, downloads the dataset from the internet and puts
        it in root directories. If dataset is already downloaded,
        it is not downloaded again.

    .. note::
        On the test subset `load_stereo` and `load_previous` can not be used
    """

    def __init__(
        self,
        kitti_raw_root: str,
        depth_prediction_root: str,
        subset: Literal["train", "val", "test"] = "train",
        load_stereo: bool = False,
        load_previous: Union[Tuple[int, int], int] = 0,
        transform: Callable[[Dict], Dict] = lambda x: x,
        download: bool = False,
    ):
        super().__init__(
            kitti_raw_root,
            depth_prediction_root,
            subset,
            load_stereo,
            load_previous,
            transform,
            download,
        )

        for path in self._paths:
            del path["lidar"]
